"""Submodel for people.py"""
from typing import Optional

from pydantic import BaseModel

from shikithon.models.anime import Anime
from shikithon.models.manga import Manga


class PeopleWorks(BaseModel):
    """Represents works entity of person."""
    anime: Optional[Anime]
    manga: Optional[Manga]
    role: str



"""Model for /api/clubs/:id/images"""
from typing import Optional

from pydantic import BaseModel


class ClubImage(BaseModel):
    """Represents club image entity."""
    id: int
    original_url: str
    main_url: str
    preview_url: str
    can_destroy: Optional[bool]
    user_id: int



"""Model for api/dialogs"""
from pydantic import BaseModel

from shikithon.models.message import Message
from shikithon.models.user import User


class Dialog(BaseModel):
    """Represents dialog entity."""
    target_user: User
    message: Message



"""Enums for /api/favorites."""
from enum import Enum


class LinkedType(Enum):
    """Contains constants related for favorite linked type."""
    ANIME = 'Anime'
    MANGA = 'Manga'
    RANOBE = 'Ranobe'
    PERSON = 'Person'
    CHARACTER = 'Character'



"""Model for /api/users/:id/favorites"""
from typing import List, Optional

from pydantic import BaseModel

from shikithon.models.favourite import Favourite


class Favourites(BaseModel):
    """Represents collection of favourites by category."""
    animes: List[Optional[Favourite]]
    mangas: List[Optional[Favourite]]
    characters: List[Optional[Favourite]]
    people: List[Optional[Favourite]]
    mangakas: List[Optional[Favourite]]
    seyu: List[Optional[Favourite]]
    producers: List[Optional[Favourite]]



"""Submodel for character.py"""
from pydantic import BaseModel

from shikithon.models.image import Image


class Seyu(BaseModel):
    """Represents seyu of character entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str



"""
Set of utility methods for the shikithon library.

This file contains the Utils class
with all the necessary utility methods
to work with the library
"""
from enum import Enum
from time import time
from typing import Any, Dict, List, Optional, Union

from loguru import logger

LOWER_LIMIT_NUMBER = 1


class Utils:
    """
    Utils class.

    Contains all the necessary utility methods
    to work with the library
    """

    @staticmethod
    def prepare_query_dict(query_dict: Dict[str, str]) -> str:
        """
        Convert query dict to query string for endpoint link.

        :param query_dict: Query dictionary
        :type query_dict: Dict[str, str]

        :return: Query string
        :rtype: str
        """
        logger.debug(f'Converting {query_dict=} to string')
        query_dict_str = '&'.join(
            f'{key}={val}' for (key, val) in query_dict.items())
        logger.debug(f'Formed {query_dict_str=}')
        return f'?{query_dict_str}'

    @staticmethod
    def convert_app_name(app_name: str) -> str:
        """
        Converts app name to snake case for use in
        config cache filename.

        :param app_name: Current OAuth app name
        :type app_name: str

        :return: Converted app name for filename
        :rtype: str
        """
        logger.debug(f'Converting {app_name=} for cached config')
        return '_'.join(app_name.lower().split(' '))

    @staticmethod
    def get_new_expire_time(time_expire_constant: int) -> int:
        """
        Generates new token expire time.

        :param time_expire_constant: Token lifetime value
        :type time_expire_constant: int

        :return: New token expire time
        :rtype: int
        """
        logger.debug(f'Getting new expiration time with a lifetime '
                     f'of {time_expire_constant} seconds')
        return int(time()) + time_expire_constant

    @staticmethod
    def generate_query_dict(
        **params_data: Optional[Union[str, bool, int, Enum, List[Union[int,
                                                                       str]]]]
    ) -> Dict[str, str]:
        """
        Returns valid query dict for API requests.

        This methods checks for data type and converts to valid one.

        :param params_data: API methods parameters data
        :type params_data:
            Optional[Union[str, bool, int, Enum, List[Union[int, str]]]]

        :return: Valid query dictionary
        :rtype: Dict[str, str]
        """
        logger.debug(
            f'Generating query dictionary for request. Passed {params_data=}')
        query_dict: Dict[str, str] = {}
        for key, data in params_data.items():
            if data is None:
                continue
            if isinstance(data, bool):
                if data is True:
                    query_dict[key] = str(int(data))
                else:
                    continue
            elif isinstance(data, int):
                query_dict[key] = str(data)
            elif isinstance(data, Enum):
                query_dict[key] = data.value
            elif isinstance(data, list):
                formatted_data: List[str] = []
                for item in data:
                    if isinstance(item, Enum):
                        formatted_data.append(item.value)
                    elif isinstance(item, int):
                        formatted_data.append(str(item))
                    elif isinstance(item, str) and item.isdigit():
                        formatted_data.append(item)
                query_dict[key] = ','.join(formatted_data)
            else:
                query_dict[key] = data
        logger.debug(f'Generated query dictionary: {query_dict=}')
        return query_dict

    @staticmethod
    def generate_data_dict(
        **dict_data: Optional[Union[str, bool, int, Enum, List[int]]]
    ) -> Union[Dict[str, str], Dict[str, Dict[str, str]]]:
        """
        Returns valid data dict for API requests.

        This methods checks for data type and converts to valid one.

        :param dict_data: API methods body data
        :type dict_data: Optional[Union[str, bool, int, Enum, List[int]]]

        :return: Valid data dictionary
        :rtype: Optional[Union[str, bool, int, Enum, List[int]]]
        """
        logger.debug(
            f'Generating data dictionary for request. Passed {dict_data=}')
        if 'dict_name' not in dict_data:
            logger.debug(
                'There is no dict_name in dict_data. Returning empty dictionary'
            )
            return {}

        logger.debug('Extracting root dictionary name')
        data_dict_name: str = dict_data['dict_name']
        dict_data.pop('dict_name')

        logger.debug(f'Setting root dictionary "{data_dict_name}"')
        new_data_dict: Dict[str, Dict[str, Any]] = {data_dict_name: {}}
        for key, data in dict_data.items():
            if data is None:
                continue
            if isinstance(data, bool):
                new_data_dict[data_dict_name][key] = data
            elif isinstance(data, int):
                new_data_dict[data_dict_name][key] = str(data)
            elif isinstance(data, Enum):
                new_data_dict[data_dict_name][key] = data.value
            elif isinstance(data, list):
                formatted_data: List[str] = []
                for item in data:
                    if isinstance(item, Enum):
                        formatted_data.append(item.value)
                    elif isinstance(item, int):
                        formatted_data.append(str(item))
                    elif isinstance(item, str) and item.isdigit():
                        formatted_data.append(item)
                new_data_dict[data_dict_name][key] = ','.join(formatted_data)
            else:
                new_data_dict[data_dict_name][key] = data
        logger.debug(f'Generated data dictionary: {new_data_dict=}')
        return new_data_dict

    @staticmethod
    def validate_query_number(query_number: Optional[int],
                              upper_limit: int) -> Optional[int]:
        """
        Validates query number.

        If number is lower, returns lower limit, else upper limit.
        If number is None, returns or None.

        :param query_number: Number to validate
        :type query_number: Optional[int]

        :param upper_limit: Upper limit for range check
        :type upper_limit: int

        :return: Validated number
        :rtype: Optional[int]
        """
        logger.debug(f'Validating query number ("{query_number}") '
                     f'with upper limit ("{upper_limit}")')
        if query_number is None:
            logger.debug('Query number is "None"')
            return query_number

        if query_number < LOWER_LIMIT_NUMBER:
            logger.debug(f'Query number ("{query_number}") is lower '
                         f'than lower limit ("{LOWER_LIMIT_NUMBER}"). '
                         f'Returning {LOWER_LIMIT_NUMBER=}')
            return LOWER_LIMIT_NUMBER

        if query_number > upper_limit:
            logger.debug(f'Query number ("{query_number}") is higher '
                         f'than upper limit ("{upper_limit}"). '
                         f'Returning {upper_limit=}')
            return upper_limit

        logger.debug(f'Returning passed query number ("{query_number}")')
        return query_number

    @staticmethod
    def query_numbers_validator(**query_numbers: List[Optional[int]]
                               ) -> Dict[str, Optional[int]]:
        """
        Gets all query numbers to validate and returns validated numbers.

        This method uses validate_query_number method for validating.

        Query numbers are passed in such form:
            { "page": [1, 100], ... }

            "page" <- Name of query number

            [1 (Passed value), 100 (Upper limit value)]

        This method outputs them like this:
            { "page": 1 }

            "page" <- Name of query number

            1 <- Validated number

        :param query_numbers: Passed query numbers to validate
        :type query_numbers: List[Optional[int]]
        :return: Dict of validated numbers
        :rtype: Dict[str, Optional[int]]
        """
        validated_numbers: Dict[str, Optional[int]] = {}
        for name, data in query_numbers.items():
            logger.debug(f'Checking "{name}" parameter')
            validated_numbers[name] = (Utils.validate_query_number(
                data[0], data[1]))
        return validated_numbers



"""Submodel for message.py"""
from typing import Optional

from pydantic import BaseModel

from shikithon.models.image import Image


class LinkedTopic(BaseModel):
    """Represents linked topic of message entity."""
    id: int
    topic_url: str
    thread_id: int
    topic_id: int
    type: str
    name: str
    russian: str
    image: Image
    url: str
    kind: str
    score: float
    status: str
    episodes: int
    episodes_aired: int
    aired_on: Optional[str]
    released_on: Optional[str]



"""
Enums for /api/person.

Also PersonKind enum is use in /api/favorites,
when linked_type is Person
"""
from enum import Enum


class PersonKind(Enum):
    """Contains constants related for favorite person kind."""
    NONE = ''
    COMMON = 'common'
    SEYU = 'seyu'
    MANGAKA = 'mangaka'
    PRODUCER = 'producer'
    PERSON = 'person'



"""Submodel for ban.py and model for /api/comments"""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from shikithon.models.user import User


class Comment(BaseModel):
    """Represents a comment entity."""
    id: int
    user_id: int
    commentable_id: int
    commentable_type: str
    body: str
    html_body: Optional[str]
    created_at: datetime
    updated_at: datetime
    is_offtopic: bool
    is_summary: Optional[bool]
    can_be_edited: Optional[bool]
    user: Optional[User]



"""Model for /api/ranobe"""
from typing import List, Optional

from pydantic import BaseModel

from shikithon.models.genre import Genre
from shikithon.models.image import Image
from shikithon.models.user_rate import UserRate
from shikithon.models.user_rate_score import UserRateScore
from shikithon.models.user_rate_status import UserRateStatus


class Ranobe(BaseModel):
    """Represents ranobe entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    kind: str
    score: float
    status: str
    volumes: int
    chapters: int
    aired_on: Optional[str]
    released_on: Optional[str]
    english: Optional[List[str]]
    japanese: Optional[List[str]]
    synonyms: Optional[List[str]]
    license_name_ru: Optional[str]
    description: Optional[str]
    description_html: Optional[str]
    description_source: Optional[str]
    franchise: Optional[str]
    favoured: Optional[bool]
    anons: Optional[bool]
    ongoing: Optional[bool]
    thread_id: Optional[int]
    topic_id: Optional[int]
    myanimelist_id: Optional[int]
    rates_scores_stats: Optional[List[UserRateScore]]
    rates_statuses_stats: Optional[List[UserRateStatus]]
    licensors: Optional[List[str]]
    genres: Optional[List[Genre]]
    publishers: Optional[List[str]]
    user_rate: Optional[UserRate]



"""Model for /api/users/:id/history"""
from datetime import datetime
from typing import Union

from pydantic import BaseModel

from shikithon.models.anime import Anime
from shikithon.models.manga import Manga


class History(BaseModel):
    """Represents user history timeline entity."""
    id: int
    created_at: datetime
    description: str
    target: Union[Anime, Manga]



"""Submodel for people.py"""
from typing import List, Optional

from pydantic import BaseModel

from shikithon.models.anime import Anime
from shikithon.models.character import Character


class PeopleRoles(BaseModel):
    """Represents roles entity of person."""
    characters: Optional[List[Character]]
    anime: Optional[List[Anime]]



"""Model for /api/users"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from shikithon.models.stats import Stats
from shikithon.models.user_image import UserImage


class User(BaseModel):
    """Represents user entity."""
    id: int
    nickname: str
    avatar: str
    image: UserImage
    last_online_at: datetime
    url: Optional[str]
    name: Optional[str]
    sex: Optional[str]
    full_years: Optional[int]
    locale: Optional[str]
    last_online: Optional[str]
    website: Optional[str]
    birth_on: Optional[datetime]
    location: Optional[str]
    banned: Optional[bool]
    about: Optional[str]
    about_html: Optional[str]
    common_info: Optional[List[str]]
    show_comments: Optional[bool]
    in_friends: Optional[bool]
    is_ignored: Optional[str]
    stats: Optional[Stats]
    style_id: Optional[int]



"""Submodel for franchise_tree.py"""
from pydantic import BaseModel


class TreeNode(BaseModel):
    """Represents tree node entity."""
    id: int
    date: int
    name: str
    image_url: str
    url: str
    year: int
    kind: str
    weight: int



"""Model for /api/achievements"""
from datetime import datetime

from pydantic import BaseModel


class Achievement(BaseModel):
    """Represents user achievement entity."""
    id: int
    neko_id: str
    level: int
    progress: int
    user_id: int
    created_at: datetime
    updated_at: datetime



"""Custom decorators for API class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from shikithon.api import API


def protected_method(scope=None):
    """
    Decorator for protected API methods.

    This method is used for all protected methods
    and checks the expiration of the access token.

    When the access token is no longer valid,
    triggers the token update function.

    Also, this decorator take a scope parameter
    for checking, if current app is allowed to access
    protected method.
    """

    def decorator(function):

        def wrapper(api: API, *args, **kwargs):
            """
            Decorator's wrapper function.

            Check for token expire time.
            If needed, triggers token refresh function.

            :return: None if API object is in restricted mode
                or if required scope is missing
            :rtype: None
            """
            logger.debug('Checking the possibility of using a protected method')
            if api.restricted_mode:
                logger.debug('It is not possible to use the protected method '
                             'due to the restricted mode')
                return None

            if scope and scope not in api.scopes_list:
                logger.debug(f'Protected method cannot be used due to the '
                             f'absence of "{scope}" scope')
                return None

            if api.token_expired():
                logger.debug('Token has expired. Refreshing...')
                api.refresh_tokens()
            logger.debug('All checks for use of the protected '
                         'method have been passed')
            return function(api, *args, **kwargs)

        return wrapper

    return decorator



"""Submodel for stats.py"""
from typing import List, Optional

from pydantic import BaseModel

from shikithon.models.type import Type


class TypeList(BaseModel):
    """Represents types collection of anime/manga."""
    anime: Optional[List[Type]]
    manga: Optional[List[Type]]



"""Submodel for creator.py"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from shikithon.models.anime import Anime
from shikithon.models.image import Image
from shikithon.models.manga import Manga
from shikithon.models.seyu import Seyu


class Character(BaseModel):
    """Represents character of anime."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    altname: Optional[str]
    japanese: Optional[str]
    description: Optional[str]
    description_html: Optional[str]
    description_source: Optional[str]
    favoured: Optional[bool]
    thread_id: Optional[int]
    topic_id: Optional[int]
    updated_at: Optional[datetime]
    seyu: Optional[List[Seyu]]
    animes: Optional[List[Anime]]
    mangas: Optional[List[Manga]]



"""Model for /api/animes/:id/topics"""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from shikithon.models.anime import Anime
from shikithon.models.forum import Forum
from shikithon.models.manga import Manga
from shikithon.models.user import User


class Topic(BaseModel):
    """Represents topic entity."""
    id: int
    topic_title: str
    body: str
    html_body: str
    html_footer: str
    created_at: datetime
    comments_count: int
    forum: Forum
    user: User
    type: str
    linked_id: int
    linked_type: str
    linked: Union[Anime, Manga]
    viewed: bool
    last_comment_viewed: Optional[bool]
    event: Optional[str]
    episode: Optional[int]



"""Submodel for favourites.py"""
from typing import Optional

from pydantic import BaseModel


class Favourite(BaseModel):
    """Represents favourite entity."""
    id: int
    name: str
    russian: str
    image: str
    url: Optional[str]



"""
Config caching module.

This module handles saving API config to cache file
to restore on next object initializaion
"""

from json import dumps, loads
from os import remove
from os.path import exists
from typing import Any, Dict, Optional, Tuple

from loguru import logger

from shikithon.utils import Utils


class ConfigCache:
    """
    Config caching class.

    This class has several methods for
    saving/restoring config from a cache file.

    In addition, there is a method
    to check if config cache file is exists.

    On saving, class takes ".shikithon_" and
    combining with the current app_name
    for generating cache file name.
    This allows to use multiple configs without
    getting new authorization codes for old configs.

    This class was created due
    to the behavior of Shikithon OAuth when
    the server refuses to use the authentication code
    when there was a second attempt to get new access tokens,
    not refreshing them
    """

    @staticmethod
    def config_name(app_name: str) -> str:
        """
        Returns name of config file for selected app name.

        :param app_name: Selected OAuth app name
        :type app_name: str

        :return: Config filename
        :rtype: str
        """
        logger.debug('Generating config cache filename')
        return '.shikithon_' + Utils.convert_app_name(app_name)

    @staticmethod
    def cache_config_validation(
            config_app_name: str,
            config_auth_code: str) -> Tuple[Dict[str, Any], bool]:
        """
        Gets cached config and validates it.

        :param config_app_name: Current app name
        :type config_app_name: str

        :param config_auth_code: Current authorization code
        :type config_auth_code: str

        :return: Tuple with config and validation status.
            On successful validation, returns cached config and True,
            otherwise, empty dictionary and False
        :rtype: Tuple[Dict[str, Any], bool]
        """
        cache_config = ConfigCache.get_config(config_app_name)
        if cache_config is None:
            logger.debug('There are no cached config')
            return {}, False

        logger.debug('Found cached config. Checking...')

        if config_auth_code and cache_config['auth_code'] != config_auth_code:
            logger.debug('Mismatch of provided and cached auth codes. '
                         'Deleting old cached config')
            ConfigCache.delete_config(config_app_name)
            return {}, False
        logger.debug('Cached config is valid')
        return cache_config, True

    @staticmethod
    def get_config(app_name: str) -> Optional[Dict[str, str]]:
        """
        Returns current config from cache file.

        :param app_name: App name for config load
        :type app_name: str

        :return: Config dictionary or None
        :rtype: Optional[Dict[str, str]]
        """
        config_name = ConfigCache.config_name(app_name)
        logger.debug(f'Getting "{config_name}" config')

        if exists(config_name):
            with open(config_name, 'r', encoding='utf-8') as config_file:
                config: Dict[str, str] = loads(config_file.read())
            return config

        logger.warning('Cache file doesn\'t exist. Returning None')
        return None

    @staticmethod
    def save_config(config: Dict[str, str]) -> bool:
        """
        Creates new cache file and saves current config.

        :param config: Current config dictionary
        :type config: Dict[str, str]

        :return: True if save succeeded, False otherwise
        :rtype: bool
        """
        config_name = ConfigCache.config_name(config['app_name'])
        logger.debug(f'Saving config to "{config_name}"')

        try:
            with open(config_name, 'w', encoding='utf-8') as config_file:
                config_file.write(dumps(config, indent=4))
            return True
        except IOError as err:
            logger.warning(f'Couldn\'t save config to file: {err}')
            return False

    @staticmethod
    def delete_config(app_name: str) -> bool:
        """
        Deletes current config file.

        :param app_name: OAuth app name
        :type app_name: str

        :return: True if delete succeeded, False otherwise
        :rtype: bool
        """
        config_name = ConfigCache.config_name(app_name)
        logger.debug(f'Deleting config "{config_name}"')

        try:
            remove(config_name)
            return True
        except OSError as err:
            logger.warning(f"Couldn't delete config: {err}")
            return False



"""Submodel for user.py"""
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field

from shikithon.models.activity import Activity
from shikithon.models.rating_list import RatingList
from shikithon.models.score_list import ScoreList
from shikithon.models.status_list import StatusList
from shikithon.models.type_list import TypeList


class Stats(BaseModel):
    """Represents user's stats entity."""
    statuses: Optional[StatusList]
    full_statuses: Optional[StatusList]
    scores: Optional[ScoreList]
    types: Optional[TypeList]
    ratings: Optional[RatingList]
    has_anime: Optional[bool] = Field(alias='has_anime?')
    has_manga: Optional[bool] = Field(alias='has_manga?')
    genres: Optional[List[str]]
    studios: Optional[List[str]]
    publishers: Optional[List[str]]
    activity: Optional[Union[List[Activity], Dict]]
